/*
 * Copyright (c) 2014,2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Jun 25 2014
 */
#include <string.h>
#include <stdlib.h>

#include <tcl.h>

/* --- local data ---------------------------------------------------------- */

#define json_skip_whitespace(json) (json) += strspn(json, " \t\n\r")


/* --- j2d_tokerror -------------------------------------------------------- */

/** Append an error message with the current token
 */
static void *
j2d_tokerror(Tcl_Interp *interp, const char *what, const char *json)
{
  char tok[128];
  size_t s;

  s = strcspn(json, " \t\n\r{}[],:\"");
  if (!s) s = 1;
  if (s > sizeof(tok)-1) s = sizeof(tok)-1;
  strncpy(tok, json, s);

  Tcl_AppendResult(interp, "unexpected ", what, " \"", tok, "\" ", NULL);
  return NULL;
}


/* --- j2d_value ----------------------------------------------------------- */

/** Parse a JSON value
 */
static Tcl_Obj *
j2d_value(Tcl_Interp *interp, const char **json)
{
  static const char tok_null[] = "null";
  static const char tok_true[] = "true";
  static const char tok_false[] = "false";

  json_skip_whitespace(*json);
  switch(**json) {
    case 0:
      return NULL;

    case '{': /* object */ {
      Tcl_Obj *dict = Tcl_NewDictObj(/*empty*/);
      Tcl_Obj *k, *v;

      ++(*json);
      json_skip_whitespace(*json);
      while(**json != '}') {
        /* key, string */
        if (**json != '"')
          return j2d_tokerror(interp, "non-string key", *json);
        if (!(k = j2d_value(interp, json))) return NULL;

        json_skip_whitespace(*json);
        if (**json != ':')
          return j2d_tokerror(interp, "element separator", *json);
        ++(*json);

        /* value */
        if (!(v = j2d_value(interp, json))) return NULL;
        if (Tcl_DictObjPut(interp, dict, k, v) != TCL_OK) return NULL;

        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        } else if (**json != '}')
          return j2d_tokerror(interp, "element", *json);
      }
      ++(*json);

      return dict;
    }

    case '[': /* array */ {
      Tcl_Obj *dict = Tcl_NewDictObj(/*empty*/);
      Tcl_Obj *k, *v;
      size_t i;

      ++(*json);
      json_skip_whitespace(*json);
      i = 0;
      while(**json != ']') {
        if (!(v = j2d_value(interp, json))) return NULL;
        if (!(k = Tcl_NewIntObj(i))) return NULL;
        if (Tcl_DictObjPut(interp, dict, k, v) != TCL_OK) return NULL;

        json_skip_whitespace(*json);
        if (**json == ',') {
          (*json)++;
          json_skip_whitespace(*json);
        } else if (**json != ']')
          return j2d_tokerror(interp, "element", *json);

        i++;
      }
      ++(*json);

      return dict;
    }

    case '"': /* string */ {
      const char *start = ++(*json);
      int quote = 0;
      char *new, *p;
      Tcl_Obj *v;

      while (**json != '"' && **json) {
        if (**json == '\\') {
          if (!*((*json)++)) break;
          if (**json == '"') quote = 1;
        }
        (*json)++;
      }
      if (!**json) return j2d_tokerror(interp, "end of string", start);
      (*json)++;
      if (!quote) return Tcl_NewStringObj(start, *json - start - 1);

      new = malloc(*json - start);
      if (!new) return NULL;
      strncpy(new, start, *json - start - 1);
      new[*json - start - 1] = 0;

      p = new;
      while((p = strstr(p, "\\"))) {
        if (*p == '"' || *p == '\\')
          memmove(p, p+1, strlen(p)/*including final \0*/);
        p++;
      }
      v = Tcl_NewStringObj(new, -1);
      free(new);
      return v;
    }

    case 't': /* bare word: true, false, null */
    case 'f':
    case 'n':
      if (!strncmp(*json, tok_true, sizeof(tok_true)-1)) {
        *json += sizeof(tok_true)-1;
        return Tcl_NewStringObj(tok_true, -1);
      }
      if (!strncmp(*json, tok_false, sizeof(tok_false)-1)) {
        *json += sizeof(tok_false)-1;
        return Tcl_NewStringObj(tok_false, -1);
      }
      if (!strncmp(*json, tok_null, sizeof(tok_null)-1)) {
        *json += sizeof(tok_null)-1;
        return Tcl_NewDictObj(/*empty*/);
      }

      return j2d_tokerror(interp, "value", *json);

    default: /* number or error */ {
      const char *start = *json;
      char *end;

      strtod(start, &end);
      if (end == start) return j2d_tokerror(interp, "value", start);
      *json = end;

      return Tcl_NewStringObj(start, *json - start);
    }
  }

  Tcl_AppendResult(interp, "internal error", NULL);
  return NULL;
}


/* --- genomix::json::j2d -------------------------------------------------- */

/** Implements the command converting a json string to a dictionary
 */
static int
j2d_cmd(ClientData d, Tcl_Interp *interp, int objc, Tcl_Obj *const objv[])
{
  const char *json;
  Tcl_Obj *v;

  if (objc != 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "json");
    return TCL_ERROR;
  }
  json = Tcl_GetString(objv[1]);

  v = j2d_value(interp, &json);
  if (!v) return TCL_ERROR;
  if (*json) {
    j2d_tokerror(interp, "extra characters at", json);
    return TCL_ERROR;
  }

  Tcl_SetObjResult(interp, v);
  return TCL_OK;
}

int
Json_Init(Tcl_Interp *interp)
{
  if (!Tcl_CreateObjCommand(
        interp, "genomix::json::json2dict", j2d_cmd, NULL, NULL))
    return TCL_ERROR;

  return TCL_OK;
}
